#ifndef V1190_IO_HH
#define V1190_IO_HH
#include <string>
#include <memory>
#include <stdint.h>
#include "ds50daq/DAQ/detail/caen_io.hh"

namespace detail {
  namespace V1190 {
    enum acquisition_modes {
      trigger_matching,
      continous_storage,
    };
    enum edge_resolutions {
      ps800 = 0,
      ps200 = 1,
      ps100 = 2,
    };
    enum edge_detections {
      pair_mode = 0,
      trailing_only = 1,
      leading_only = 2,
      trailing_and_leading = 3,
    };
    enum status_masks {
      dready_msk = 1,
      almost_full_msk = 2,
      full_msk = 4,
      trg_match_msk = 8,
      header_en_msk = 0x10,
      term_on_msk = 0x20,
      error_0_msk = 0x40,
      error_1_msk = 0x80,
      error_2_msk = 0x100,
      error_3_msk = 0x200,
      error_any_msk = 0x3C0,
      berr_flag_msk = 0x400,
      purged_msk = 0x800,
      res_0_msk = 0x1000,
      res_1_msk = 0x2000,
      pair_mode_msk = 0x4000,
      trigger_lost_msk = 0x8000,
    };
    enum control_masks {
      berr_enable_msk = 1,
      term_msk = 2,
      term_sw_msk = 4,
      empty_event_msk = 8,
      align64_msk = 0x10,
      compensation_enable_msk = 0x20,
      test_fifo_enable_msk = 0x40,
      read_compensation_sram_enable_msk= 0x80,
      event_fifo_enable_msk = 0x100,
      trigger_time_tag_enable_msk = 0x200,
    };
    enum uc_opcodes {
      trg_match = 0,
      cont_store = 0x100,
      read_acq_mod = 0x200,

      set_win_width = 0x1000,
      set_win_offset = 0x1100,
      en_sub_trg = 0x1400,
      dis_sub_trg = 0x1500,

      set_detection = 0x2200,
      read_detection = 0x2300,
      set_tr_lead_lsb = 0x2400,
      set_pair_res = 0x2500,
      read_res = 0x2600,
      set_dead_time = 0x2800,

      en_head_trailer = 0x3000,
      dis_head_trailer = 0x3100,
      set_event_size = 0x3300,
      read_event_size = 0x3400,

      en_channel = 0x4000,
      dis_channel = 0x4100,
      write_en_pattern = 0x4400,

      read_tdc_id = 0x6000,
    };
    enum micro_codes_msk {
      write_ok_msk = 1,
      read_ok_msk = 2,
    };
    enum registers {
      output = 0,
      control = 0x1000,
      status = 0x1002,
      int_level = 0x100a,
      int_vector = 0x100c,
      geo_address = 0x100e,
      mod_reset = 0x1014,
      sw_clear = 0x1016,
      sw_event_reset = 0x1018,
      sw_trigger = 0x101a,
      event_counter_d32 = 0x101c,
      event_stored = 0x1020,
      almost_full_lvl = 0x1022,
      blt_event_num = 0x1024,
      fw_rev = 0x1026,
      out_prog_ctrl = 0x102c,
      micro = 0x102e,
      micro_hnd = 0x1030,
      event_fifo_d32 = 0x1038,
      event_fifo_stored = 0x103c,
      event_fifo_status = 0x103e,
      oui0 = 0x402C,
      oui1 = 0x4028,
      oui2 = 0x4024
    };
  }

  class V1190_io {
    public:
      V1190_io (uint32_t vme_base, const std::string& link_type, int link_number, int link_node);

      void clear ();
      void soft_trigger ();

      uint16_t status ();
      bool data_ready () { return status () & V1190::dready_msk; }
      bool full () { return status () & V1190::full_msk; }
      bool error () { return status () & V1190::error_any_msk; }

      uint16_t control ();
      void empty_events (bool on);
      void termination (bool on);
      void trigger_time_tag (bool on);

      void geo (int id);
      uint16_t events_stored ();
      uint16_t events_counter ();

      void acquisition_mode (V1190::acquisition_modes);
      V1190::acquisition_modes acquisition_mode ();

      void edge_resolution (V1190::edge_resolutions);

      void edge_detection (V1190::edge_detections);
      V1190::edge_detections edge_detection ();

      void matching_window_width_25ns (u_int16_t);
      void matching_window_offset_25ns (int16_t);

      void tdc_header (bool);
      void trigget_time_subtraction (bool);

      bool wait_irq (uint32_t timeout_ms);
      ssize_t read_event (void* destination, size_t max_size);

    private:
      std::unique_ptr<caen_io> caen_io_;
      uint8_t irq_level_;

      void wait_uc_ (V1190::micro_codes_msk mode, int timeout_cycles = 100);
      void write_uc_ (u_int16_t data);
      u_int16_t read_uc_ ();
      void send_uc_ (V1190::uc_opcodes opcode, u_int16_t data);
      u_int16_t recv_uc_ (V1190::uc_opcodes opcode);

      void enable_irq_ (uint8_t level = 1, uint8_t vector = 0x3e);
      void reset_ ();
  };
}

#endif
