#include "ds50daq/DAQ/V1190Driver.hh"

#include "artdaq/DAQdata/Debug.hh"
#include "artdaq/DAQdata/GeneratorMacros.hh"
#include "cetlib/exception.h"
#include "ds50daq/DAQ/Config.hh"
#include "ds50daq/DAQ/V1190Fragment.hh"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "detail/V1190_io.hh"

#include <string>
#include <string.h>
#include <vector>

using fhicl::ParameterSet;
using ds50::V1190Fragment;
namespace V1190 = detail::V1190;

ds50::V1190Driver::V1190Driver(ParameterSet const & ps): DS50FragmentGenerator(ps.get<ParameterSet> ("generator_ds50")) {
  init_link (ps);

  init_trigger (ps);

  init_channels (ps);
}

ds50::V1190Driver::~V1190Driver() noexcept {
  close_link ();
}

void ds50::V1190Driver::close_link () {
  v1190_ = 0;
}


void ds50::V1190Driver::init_link (ParameterSet const& ps) {
  v1190_ = std::unique_ptr<detail::V1190_io>(new detail::V1190_io (ps.get<uint32_t> ("vme_base"), ps.get<std::string> ("link_type"), ps.get<int> ("link_number"), ps.get<int> ("link_node", 0)));

  v1190_->termination (ps.get<bool> ("control_lines_termination", true));
  v1190_->empty_events (true);

  irqwait_ms_ = ps.get<int> ("irqwait_ms", 200);
  irqwait_checks_ = ps.get<int> ("irqwait_checks", 2);

  io_buffer_length_ = ps.get<int> ("io_buffer_length", 10000);
  io_buffer_ = std::unique_ptr<uint32_t[]>(new uint32_t[io_buffer_length_]);
}

void ds50::V1190Driver::init_trigger (ParameterSet const& ps) {
  v1190_->acquisition_mode (V1190::trigger_matching);
  v1190_->edge_detection (V1190::trailing_and_leading);
  v1190_->edge_resolution (V1190::ps800);
  v1190_->matching_window_width_25ns (int(ps.get<float> ("acquisition_gate_us") * 40));
  v1190_->matching_window_offset_25ns (int16_t (ps.get<float> ("pre_trigger_us") * 40));


  v1190_->trigger_time_tag (ps.get<bool> ("trigger_time_tag", false));
  v1190_->tdc_header (ps.get<bool> ("tdc_header", false));
  v1190_->trigget_time_subtraction (ps.get<bool> ("trigget_time_subtraction", true));
}

void ds50::V1190Driver::init_channels (ParameterSet const&) {
}

void ds50::V1190Driver::start_ () {
  v1190_->clear ();
}

void ds50::V1190Driver::stop_ () {
}


bool ds50::V1190Driver::getNext__(artdaq::FragmentPtrs& frags) {
  bool will_stop = should_stop ();

  //v1190_->soft_trigger ();
  for (int j = 0; j < irqwait_checks_; j++) {
    if (!v1190_->data_ready () && !v1190_->wait_irq (irqwait_ms_)) continue;
    //if (!v1190_->wait_irq (irqwait_ms_)) continue;

    ssize_t event_size = v1190_->read_event (io_buffer_.get (), io_buffer_length_ * sizeof (io_buffer_[0]));
    if (event_size == 0) continue;
    else if (event_size < 0) event_size = -event_size;

    std::unique_ptr<artdaq::Fragment> frag(new artdaq::Fragment(std::ceil (event_size / static_cast<double>(sizeof (artdaq::RawDataType)))));
    memcpy (&*frag->dataBegin (), io_buffer_.get (), event_size);
    V1190Fragment const v1190_frag(*frag.get ());

    //std::cout << std::hex << std::showbase << v1190_frag << std::endl;
    if (v1190_frag.dataBegin ()[0].type () != ds50::V1190Fragment::Word::global_header)
      throw cet::exception("V1190DriverGenerator") << "first word is not a global header";

    int ev_counter_diff = v1190_frag.dataBegin ()[0].event_count () + 1 - ev_counter ();
    if (ev_counter_diff) {
      mf::LogError ("V1190") << "board " << fragment_id () << " jump " << ev_counter_diff << "(" << v1190_frag.dataBegin ()[0].event_count () << " vs " << ev_counter () << ") at event " << ev_counter ();

      for (int i = 0; ev_counter_diff < 100 && i < ev_counter_diff; ++i)
        frags.emplace_back(new artdaq::Fragment(ev_counter () + i, fragment_id (), Config::MISSED_FRAGMENT_TYPE));

      ev_counter_inc (ev_counter_diff);
    }

    frag->setFragmentID (fragment_id ());
    frag->setSequenceID (ev_counter_inc ());
    frag->setUserType (Config::V1190_FRAGMENT_TYPE);

    frags.emplace_back (std::move (frag));
  }

  if (!frags.size () && will_stop) return false;
  return true;
}

DEFINE_ARTDAQ_GENERATOR(ds50::V1190Driver)
