#include "ds50daq/DAQ/V1495Driver.hh"

#include "artdaq/DAQdata/Debug.hh"
#include "artdaq/DAQdata/GeneratorMacros.hh"
#include "cetlib/exception.h"
#include "ds50daq/DAQ/Config.hh"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "detail/caen_io.hh"

#include <algorithm>
#include <string>
#include <vector>
#include <cmath>

using fhicl::ParameterSet;
using ds50::V1495Fragment;

namespace V1495 {
  enum V1495_registers {
    fw_revision = 0x100C,
    port_a_mask = 0x1010,
    port_b_mask = 0x1014,
    run_control = 0x1018,
    window_inibit = 0x101C,
    mf_extention = 0x1020,
    trigger_control = 0x1024,
    run_number = 0x1028,
    test_pulser_dpm = 0x102C,
    status = 0x1030,
    test_pulser_read = 0x1034,
    tdc_delay = 0x1044,
    laser_control = 0x1048,
    irq_latch = 0x1080,
    irq_level = 0x8004,
    irq_id = 0x8006,
    reset_module = 0x800A,
    serial_msb = 0x8180,
    serial_lsb = 0x8184
  };
  uint32_t frequency_to_pattern (float frequency) { return (std::max (0, std::min (0xFFFF, int(std::lround ((6.25e3 / frequency - 1) * 0xFFFF / (62.5e3 - 1)))))) << 16; }
}


ds50::V1495Driver::V1495Driver(ParameterSet const & ps): DS50FragmentGenerator(ps.get<ParameterSet> ("generator_ds50")) {
  if (ps.get<bool> ("reset_vme", false)) {
    init_link (ps);
    caen_io_->vme_sysreset ();
    close_link ();
    usleep (1000000);
    mf::LogDebug("V1495") << "Performed VME SYSReset";
  }

  init_link (ps);

  init_metadata (ps);

  init_trigger (ps);

  init_channels (ps);

  dont_stop_ = ps.get<bool> ("dont_stop", false);
}

ds50::V1495Driver::~V1495Driver() noexcept {
  close_link ();
}

void ds50::V1495Driver::close_link () {
  caen_io_ = 0;
}


void ds50::V1495Driver::init_link (ParameterSet const& ps) {
  caen_io_ = std::unique_ptr<detail::caen_io>(new detail::caen_io (cvA32_U_DATA, cvD32, ps.get<uint32_t> ("vme_base"), ps.get<std::string> ("link_type"), ps.get<int> ("link_number"), ps.get<int> ("link_node", 0)));

  caen_io_->write_register (V1495::reset_module, 1);
  usleep (100000);

  caen_io_->write_register (V1495::irq_level, 1);

  irqwait_ms_ = ps.get<int> ("irqwait_ms", 200);
  irqwait_checks_ = ps.get<int> ("irqwait_checks", 2);
}

void ds50::V1495Driver::init_metadata (ParameterSet const& ps) {
  metadata_.serial = ((caen_io_->read_register (V1495::serial_msb, cvD16) & 0xff) << 8) | (caen_io_->read_register (V1495::serial_lsb, cvD16) & 0xff);
  metadata_.firmware_version = caen_io_->read_register (V1495::fw_revision);

  metadata_.threshold = ps.get<int> ("low_threshold");
}

void ds50::V1495Driver::init_trigger (ParameterSet const& ps) {
  caen_io_->write_register (V1495::run_control, 1 << 1);
  caen_io_->write_register (V1495::window_inibit, int(std::max(ps.get<float> ("acquisition_gate_us"), 10.F) / 20e-3));

  if (ps.get<float> ("memory_full_extension_us", -1) > 0) {
    caen_io_->set_register_bits (V1495::run_control, 4);
    caen_io_->write_register (V1495::mf_extention, int(ps.get<float> ("memory_full_extension_us") / 20e-3));
  }

  if (metadata_.threshold) {
    int low_threshold = std::min (abs(metadata_.threshold), 39);
    int high_threshold = std::min (ps.get<unsigned int> ("high_threshold", 39), 39U);
    int majority_time_10ns = std::min (ps.get<unsigned int> ("majority_time_10ns", 10), 15U);
    caen_io_->clear_register_bits (V1495::trigger_control, 0xffff0000);  // First reset the values
    caen_io_->set_register_bits (V1495::trigger_control, 1 | (majority_time_10ns << 16) | (low_threshold << 20) | (high_threshold << 26));
    if (metadata_.threshold < 0) caen_io_->write_register (V1495::run_control, 0);
  } else caen_io_->clear_register_bits (V1495::trigger_control, 1); 

  caen_io_->set_register_bits (V1495::trigger_control, (1 << 9) | (1 << 8));  // enable external E4 and laser on E5

  bool random_triggers = ps.get<bool> ("random_triggers", false);
  if (random_triggers) caen_io_->set_register_bits (V1495::trigger_control, 1 << 2);
  else caen_io_->clear_register_bits (V1495::trigger_control, 1 << 2);

  float laser_frequency = ps.get<float> ("laser_frequency", -1);
  if (laser_frequency != 0) {
    if (laser_frequency > 0) caen_io_->write_register (V1495::laser_control, 1 | V1495::frequency_to_pattern (laser_frequency)); // No E5 input if internal mode is enabled
    else caen_io_->write_register (V1495::laser_control, 1 << 1);  // E5 enabled
    if (!metadata_.threshold) { 
      caen_io_->clear_register_bits (V1495::trigger_control, 0xffff0000);  // First reset the values
      caen_io_->set_register_bits (V1495::trigger_control, (15UL << 16) | (1UL << 20) | (39UL << 26));
    }
  } else caen_io_->write_register (V1495::laser_control, 0); // laser completelly disabled

  float pulser_frequency = ps.get<float> ("pulser_frequency", -1);
  if (pulser_frequency > 0) {
    caen_io_->set_register_bits (V1495::run_control, V1495::frequency_to_pattern (pulser_frequency));
    caen_io_->set_register_bits (V1495::trigger_control, 1 << 1);
  } else caen_io_->clear_register_bits (V1495::trigger_control, 1 << 1);

  if (ps.get<bool> ("tpm_triggers", false)) {
    for (uint8_t i = 0; i < 8; i++) {
      uint32_t reg = ((uint16_t(1) << (i + 1)) - 1);
      caen_io_->write_register (V1495::test_pulser_dpm, reg << 16);
      caen_io_->write_register (V1495::test_pulser_dpm, (i * 0x100) << 16 | 1);
    }
    caen_io_->write_register (V1495::test_pulser_dpm, 0);
    caen_io_->set_register_bits (V1495::trigger_control, 1 << 10);
  } else caen_io_->clear_register_bits (V1495::trigger_control, 1 << 10);

  caen_io_->write_register (V1495::tdc_delay, int(ps.get<float> ("tdc_pulse_delay_us", 100) / 20e-3));

  if (ps.get<bool> ("fnal_mode", false)) caen_io_->set_register_bits (V1495::run_control, 1 << 15);
}


void ds50::V1495Driver::init_channels (ParameterSet const& ps) {
  std::vector<int> masked_channels = ps.get <std::vector<int>> ("masked_channels_0bias", std::vector<int>(0));
  uint64_t mask = 0;
  for (int ch: masked_channels) {
    if (ch < 15) mask |= 1 << ch;
    else if (ch < 30) mask |= 1 << (ch + 1);
    else mask |= 1 << (ch + 2);
  }
  if (mask) {
    caen_io_->write_register (V1495::port_a_mask, ~uint32_t(mask));
    caen_io_->write_register (V1495::port_b_mask, ~uint32_t(mask >> 32) & 0x3FF);
  }
}

void ds50::V1495Driver::start_ () {
  caen_io_->write_register (V1495::run_number, run_number ());
  irqlost_ = 0;

  pause_ ();
  caen_io_->set_register_bits (V1495::run_control, 1);
  resume_ ();
}

void ds50::V1495Driver::stop_ () {
  if (!dont_stop_) {
    pause_ (); 
    caen_io_->clear_register_bits (V1495::run_control, 1);
    resume_ ();
  }
}

void ds50::V1495Driver::pause_ () {
  caen_io_->set_register_bits (V1495::run_control, 1 << 4);
  usleep (1000);
}

void ds50::V1495Driver::resume_ () {
  usleep (1000);
  caen_io_->clear_register_bits (V1495::run_control, 1 << 4);
}

void ds50::V1495Driver::shutdown_ () {
  caen_io_->vme_sysreset ();
}

bool ds50::V1495Driver::getNext__(artdaq::FragmentPtrs& frags) {
  bool will_stop = should_stop ();

  for (int j = 0; j < irqwait_checks_; j++) {
    uint32_t v;
    if (!((will_stop && !(caen_io_->read_register (V1495::status) & 0x20)) || caen_io_->wait_irq (cvIRQ1, irqwait_ms_, v, cvD8))) continue;

    if (gettimeofday (&metadata_.trigger_time, nullptr) < 0) mf::LogError ("V1495") << "gettimeofday failed";

    std::unique_ptr<artdaq::Fragment> frag(new artdaq::Fragment(std::ceil (V1495Fragment::pod_size () / static_cast<double>(sizeof (artdaq::RawDataType))), ev_counter (), fragment_id (), Config::V1495_FRAGMENT_TYPE, metadata_));

    uint32_t status_r = caen_io_->read_register (V1495::status);
    if ((status_r & 0x20)) mf::LogError ("V1495") << "data not ready (status = " << std::showbase << std::hex << status_r << ")";

    try {
      int read_size = caen_io_->read_block (0x2000, &*frag->dataBegin (), V1495Fragment::pod_size (), detail::caen_io::sequential_fifo);
      if (read_size != int(V1495Fragment::pod_size ())) mf::LogWarning ("V1495") << "Incomplete transfer " << read_size;
    } catch (...) {
      mf::LogError ("V1495") << "board " << fragment_id () << " I/O error (status = " << std::showbase << std::hex << status_r << ") at event " << std::dec << ev_counter ();
      throw;
    }

    caen_io_->write_register (V1495::irq_latch, 1);

#if DUMP_EVENTS
    std::cout << std::dec << "Ev " << ev_counter () << std::endl;
    std::cout << std::hex << std::showbase << "Run control " << caen_io_->read_register (V1495::run_control) << std::endl;
    std::cout << std::hex << std::showbase << "Trigger control " << caen_io_->read_register (V1495::trigger_control) << std::endl;
    std::cout << std::hex << std::showbase << "Status " << caen_io_->read_register (V1495::status) << std::endl;
    for (size_t i = 0; i < V1495Fragment::pod_size () / 4; i++) std::cout << std::hex << std::showbase << ((uint32_t*)&*frag->dataBegin ())[i] << std::endl;
    std::cout << "------------------------------" << std::endl;
#endif

    V1495Fragment const v1495_frag(*frag.get ());
    //if (!(ev_counter () % 4096))  std::cout << std::hex << std::showbase << v1495_frag << std::endl;

    int ev_counter_diff = v1495_frag.trigger_counter () - ev_counter ();
    if (ev_counter_diff > 0) {
      mf::LogError ("V1495") << "board " << fragment_id () << " jump " << ev_counter_diff << " (got " << v1495_frag.trigger_counter () << ") at event " << ev_counter ();

      for (int i = 0; ev_counter_diff < 100 && i < ev_counter_diff; ++i)
        frags.emplace_back(new artdaq::Fragment(ev_counter () + i, fragment_id (), Config::MISSED_FRAGMENT_TYPE));

      ev_counter_inc (ev_counter_diff);
      frag->setSequenceID (ev_counter ());
    } else if (ev_counter_diff < 0) {
      mf::LogError ("V1495") << "board " << fragment_id () << " duplicated fragment " << v1495_frag.trigger_counter () << " at event " << ev_counter ();
      continue;
    }

    frags.emplace_back (std::move (frag));
    ev_counter_inc ();
  }

  if (!frags.size () && will_stop) {
    // 16-Apr-2013, KAB:  added try/catch and single retry of the
    // status register read since it often fails the first time at
    // the Fermilab WH14 teststand.
    try {
      mf::LogDebug("V1495") << "status is " << std::showbase << std::hex << caen_io_->read_register (V1495::status);
    }
    catch (...) {
      try {
        mf::LogDebug("V1495") << "status is " << std::showbase << std::hex << caen_io_->read_register (V1495::status);
      }
      catch (...) {
        mf::LogDebug("V1495") << "status is not readable.";
      }
    }
    return false;
  }
  return true;
}

DEFINE_ARTDAQ_GENERATOR(ds50::V1495Driver)
