#include "ds50daq/DAQ/V172xDriver.hh"

#include "artdaq/DAQdata/Debug.hh"
#include "artdaq/DAQdata/GeneratorMacros.hh"
#include "cetlib/exception.h"
#include "ds50daq/DAQ/Config.hh"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "CAENDigitizer.h"
#include "detail/const_map_accessors.hh"

#include <string>
#include <vector>
#include <map>
#include <iterator>

using fhicl::ParameterSet;
using ds50::V172xFragment;
namespace cma = detail::const_map_accessors;

const std::string ds50::V172xDriver::READ_TIME_STAT_KEY("V172xDriverReadTime");
const std::string ds50::V172xDriver::FRAGMENT_CREATE_TIME_STAT_KEY("V172xDriverFragmentCreateTime");
const std::string ds50::V172xDriver::MEMCPY_TIME_STAT_KEY("V172xDriverMemcpyTime");

namespace {
  const std::string caen_error (CAEN_DGTZ_ErrorCode err) {
    static const std::map<CAEN_DGTZ_ErrorCode, std::string> errors{{CAEN_DGTZ_Success, "CAEN_DGTZ_Success" },
            { CAEN_DGTZ_CommError, "CAEN_DGTZ_CommError" },
            { CAEN_DGTZ_GenericError, "CAEN_DGTZ_GenericError" },
            { CAEN_DGTZ_InvalidParam, "CAEN_DGTZ_InvalidParam" },
            { CAEN_DGTZ_InvalidLinkType, "CAEN_DGTZ_InvalidLinkType" },
          //  { CAEN_DGTZ_InvalidHandler, "CAEN_DGTZ_InvalidHandler" },
            { CAEN_DGTZ_MaxDevicesError, "CAEN_DGTZ_MaxDevicesError" },
            { CAEN_DGTZ_BadBoardType, "CAEN_DGTZ_BadBoardType" },
            { CAEN_DGTZ_BadInterruptLev, "CAEN_DGTZ_BadInterruptLev" },
            { CAEN_DGTZ_BadEventNumber, "CAEN_DGTZ_BadEventNumber" },
            { CAEN_DGTZ_ReadDeviceRegisterFail, "CAEN_DGTZ_ReadDeviceRegisterFail" },
            { CAEN_DGTZ_WriteDeviceRegisterFail, "CAEN_DGTZ_WriteDeviceRegisterFail" },
            { CAEN_DGTZ_InvalidChannelNumber, "CAEN_DGTZ_InvalidChannelNumber" },
            { CAEN_DGTZ_ChannelBusy, "CAEN_DGTZ_ChannelBusy" },
            { CAEN_DGTZ_FPIOModeInvalid, "CAEN_DGTZ_FPIOModeInvalid" },
            { CAEN_DGTZ_WrongAcqMode, "CAEN_DGTZ_WrongAcqMode" },
            { CAEN_DGTZ_FunctionNotAllowed, "CAEN_DGTZ_FunctionNotAllowed" },
            { CAEN_DGTZ_Timeout, "CAEN_DGTZ_Timeout" },
            { CAEN_DGTZ_InvalidBuffer, "CAEN_DGTZ_InvalidBuffer" },
            { CAEN_DGTZ_EventNotFound, "CAEN_DGTZ_EventNotFound" },
            { CAEN_DGTZ_InvalidEvent, "CAEN_DGTZ_InvalidEvent" },
            { CAEN_DGTZ_OutOfMemory, "CAEN_DGTZ_OutOfMemory" },
            { CAEN_DGTZ_CalibrationError, "CAEN_DGTZ_CalibrationError" },
            { CAEN_DGTZ_DigitizerNotFound, "CAEN_DGTZ_DigitizerNotFound" },
            { CAEN_DGTZ_DigitizerAlreadyOpen, "CAEN_DGTZ_DigitizerAlreadyOpen" },
            { CAEN_DGTZ_DigitizerNotReady, "CAEN_DGTZ_DigitizerNotReady" },
            { CAEN_DGTZ_InterruptNotConfigured, "CAEN_DGTZ_InterruptNotConfigured" },
            { CAEN_DGTZ_DigitizerMemoryCorrupted, "CAEN_DGTZ_DigitizerMemoryCorrupted" },
            { CAEN_DGTZ_DPPFirmwareNotSupported, "CAEN_DGTZ_DPPFirmwareNotSupported" },
            { CAEN_DGTZ_NotYetImplemented, "CAEN_DGTZ_NotYetImplemented" }};
    auto it = errors.find (err);
    if (it == errors.end ()) return "CAEN_Unknown_error";
    return it->second;
  }

  static const std::map<std::string, CAEN_DGTZ_ConnectionType> link_types{{ "USB", CAEN_DGTZ_USB },
                                                { "PCI", CAEN_DGTZ_PCI_OpticalLink },
                                                { "PCIE", CAEN_DGTZ_PCIE_OpticalLink }};

  static const std::map<std::string, CAEN_DGTZ_IOLevel_t> io_levels{{ "NIM", CAEN_DGTZ_IOLevel_NIM },
                { "TTL", CAEN_DGTZ_IOLevel_TTL }};

  static const std::map<std::string, CAEN_DGTZ_AcqMode_t> acquisition_modes{{ "SW", CAEN_DGTZ_SW_CONTROLLED },
                { "HW", CAEN_DGTZ_S_IN_CONTROLLED }};

  static const std::map<CAEN_DGTZ_BoardFamilyCode_t, int> sample_rates_MHz{{ CAEN_DGTZ_XX724_FAMILY_CODE, 100 },
                { CAEN_DGTZ_XX721_FAMILY_CODE, 500 },
            { CAEN_DGTZ_XX731_FAMILY_CODE, 500 },
            { CAEN_DGTZ_XX720_FAMILY_CODE, 250 },
            { CAEN_DGTZ_XX740_FAMILY_CODE, 65 },
            { CAEN_DGTZ_XX751_FAMILY_CODE, 1000 },
            { CAEN_DGTZ_XX742_FAMILY_CODE, 5000 },
            { CAEN_DGTZ_XX780_FAMILY_CODE, 100 },
            { CAEN_DGTZ_XX761_FAMILY_CODE, 4000 }};

  static const std::map<CAEN_DGTZ_BoardFamilyCode_t, int> v17xx_modules{{ CAEN_DGTZ_XX724_FAMILY_CODE, 24 },
                { CAEN_DGTZ_XX721_FAMILY_CODE, 21 },
            { CAEN_DGTZ_XX731_FAMILY_CODE, 31 },
            { CAEN_DGTZ_XX720_FAMILY_CODE, 20 },
            { CAEN_DGTZ_XX740_FAMILY_CODE, 40 },
            { CAEN_DGTZ_XX751_FAMILY_CODE, 51 },
            { CAEN_DGTZ_XX742_FAMILY_CODE, 42 },
            { CAEN_DGTZ_XX780_FAMILY_CODE, 80 },
            { CAEN_DGTZ_XX761_FAMILY_CODE, 61 }};
}


ds50::V172xDriver::V172xDriver(ParameterSet const & ps): DS50FragmentGenerator(ps.get<ParameterSet> ("generator_ds50")) {
  statsHelper_.addMonitoredQuantityName(READ_TIME_STAT_KEY);
  statsHelper_.addMonitoredQuantityName(FRAGMENT_CREATE_TIME_STAT_KEY);
  statsHelper_.addMonitoredQuantityName(MEMCPY_TIME_STAT_KEY);
  statsHelper_.createCollectors(ps, 100, 30.0, 180.0);

  init_link (ps);

  init_metadata (ps);

  init_channels (ps);

  init_trigger (ps);

  init_buffers (ps);

  generate_fragment_ = ps.get<bool> ("generate_fragment", true);
  pattern_check_ = ps.get<bool> ("pattern_check", false);
}

ds50::V172xDriver::~V172xDriver() noexcept {
  close_link ();
}

void ds50::V172xDriver::close_link () {
  try {
    stop_ ();  // stop already called?
  } catch (...) {
    throw;
  }

  CAEN_DGTZ_FreeReadoutBuffer (&event_buffer_);

  CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_CloseDigitizer (handle_);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_CloseDigitizer(" << handle_ << "): " << caen_error (err);
}


void ds50::V172xDriver::init_link (ParameterSet const& ps) {
  CAEN_DGTZ_ConnectionType link_type = cma::get (link_types, ps.get<std::string> ("link_type"), "link_type");
  int link_number = ps.get<int> ("link_number");
  int link_node = ps.get<int> ("link_node", 0);
  uint32_t vme_base = ps.get<uint32_t> ("vme_base", 0UL);
  CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_OpenDigitizer (link_type, link_number, link_node, vme_base, &handle_);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_OpenDigitizer(" << cma::rget(link_types, link_type) << "," << link_number << "," << link_node << "," << vme_base << "): " << caen_error (err);
  mf::LogDebug ("V172x") << "digitizer found at (" << link_type << "," << link_number << "," << link_node << "," << vme_base << ") with handle " << handle_;

  err = CAEN_DGTZ_Reset (handle_);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_Reset(" << handle_ << "):" << caen_error (err);
  usleep (100000);

  set_register (0xEF08, fragment_id ());

  link_type_ = int(link_type);
  vme_ = ps.get<bool> ("vme", false);
}

void ds50::V172xDriver::init_metadata (ParameterSet const& ps) {
  CAEN_DGTZ_BoardInfo_t BoardInfo;

  CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_GetInfo (handle_, &BoardInfo);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_GetInfo(" << handle_ << "): " << caen_error (err);
  mf::LogDebug ("V172x") << "digitizer " << BoardInfo.SerialNumber << " model " << BoardInfo.ModelName << " ROC # " << BoardInfo.ROC_FirmwareRel << " AMC # " << BoardInfo.AMC_FirmwareRel;

  metadata_.serial_number = BoardInfo.SerialNumber;
  strncpy (metadata_.roc_firmware, BoardInfo.ROC_FirmwareRel, sizeof (metadata_.roc_firmware));
  strncpy (metadata_.amc_firmware, BoardInfo.AMC_FirmwareRel, sizeof (metadata_.amc_firmware));
  metadata_.sample_bits = BoardInfo.ADC_NBits;
  metadata_.v17xx_model = cma::get (v17xx_modules, CAEN_DGTZ_BoardFamilyCode_t(BoardInfo.FamilyCode), "v17xx_model");
  metadata_.sample_rate_MHz = cma::get (sample_rates_MHz, CAEN_DGTZ_BoardFamilyCode_t(BoardInfo.FamilyCode), "sample_rates_MHz");

  switch (BoardInfo.FamilyCode) {
    case CAEN_DGTZ_XX720_FAMILY_CODE:
      fragment_type_ = Config::V1720_FRAGMENT_TYPE;
      break;
    case CAEN_DGTZ_XX724_FAMILY_CODE:
      fragment_type_ = Config::V1724_FRAGMENT_TYPE;
      break;
    default:
      fragment_type_ = Config::MISSED_FRAGMENT_TYPE;
      mf::LogError ("V172x") << "unknown fragment type for digitizer " << int(metadata_.v17xx_model);
  }
  float acquisition_gate_us = ps.get<float> ("acquisition_gate_us");
  metadata_.record_length = int(acquisition_gate_us * metadata_.sample_rate_MHz);

  float pre_trigger_us = ps.get<float> ("pre_trigger_us");
  if (pre_trigger_us > acquisition_gate_us) throw cet::exception("Configuration") << "pre_trigger_us (" << pre_trigger_us << ") > acquisition_gate_us (" << acquisition_gate_us << ")";
  metadata_.post_trigger = std::min(100 - int((100. * pre_trigger_us) / acquisition_gate_us), 99);
}

void ds50::V172xDriver::init_channels (ParameterSet const& ps) {
  std::vector<ParameterSet> chs_ps = ps.get<std::vector<ParameterSet>>("channels");

  u_int8_t channels_mask = 0;
  for (ParameterSet ch_ps: chs_ps) {
    int enable = ch_ps.get<bool> ("enable");
    if (!enable) continue;

    int channel = ch_ps.get<int> ("channel");
    if (channel < 0 || channel > 7) throw cet::exception("Configuration") << "channel out of bounds (" << channel << ")";
    channels_mask |= 1 << channel;

    int dc_offset = ch_ps.get<int> ("dc_offset");
    if (dc_offset < 0 || dc_offset > 0xFFFF) throw cet::exception("Configuration") << "channel offset out of bounds (" << dc_offset << ")";

    CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_SetChannelDCOffset (handle_, channel, dc_offset);
    if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetChannelDCOffset(" << handle_ << "," << channel << "," << dc_offset << "): " << caen_error (err);
  }

  CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_SetChannelEnableMask (handle_, channels_mask);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetChannelEnableMask(" << handle_ << "," << channels_mask << "): " << caen_error (err);

  uint32_t reg = get_register (0x8000);
  if (reg & ((0x1111 << 16) | (1 << 11) | (1 << 6) | (1 << 1))) throw cet::exception("Configuration") << "unsupported channel status " << std::hex << std::showbase << reg;

  if (ps.get<bool> ("test_pattern", false)) set_register_bits (0x8000, 1 << 3);
}


void ds50::V172xDriver::init_trigger (ParameterSet const& ps) {
  CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_SetExtTriggerInputMode (handle_, CAEN_DGTZ_TRGMODE_ACQ_AND_EXTOUT);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetExtTriggerInputMode(" << handle_ << ",CAEN_DGTZ_TRGMODE_ACQ_AND_EXTOUT): " << caen_error (err);

  err = CAEN_DGTZ_SetSWTriggerMode(handle_, CAEN_DGTZ_TRGMODE_ACQ_AND_EXTOUT);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetSWTriggerMode(" << handle_ << ",CAEN_DGTZ_TRGMODE_ACQ_AND_EXTOUT): " << caen_error (err);

  err = CAEN_DGTZ_SetChannelSelfTrigger (handle_, CAEN_DGTZ_TRGMODE_DISABLED, 0xff);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetChannelSelfTrigger(" << handle_ << ",CAEN_DGTZ_TRGMODE_DISABLED,0xff): " << caen_error (err);

  CAEN_DGTZ_IOLevel_t io_level = cma::get (io_levels, ps.get<std::string> ("io_level", "NIM"), "io_level");
  err = CAEN_DGTZ_SetIOLevel (handle_, io_level);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetIOLevel(" << handle_ << "," << cma::rget (io_levels, io_level) << "): " << caen_error (err);

  err = CAEN_DGTZ_SetRecordLength (handle_, metadata_.record_length);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetIOLevel(" << handle_ << "," << metadata_.record_length << "): " << caen_error (err);

  int post_trigger = metadata_.post_trigger;
  err = CAEN_DGTZ_SetPostTriggerSize (handle_, post_trigger);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetPostTriggerSize(" << handle_ << "," << post_trigger << "): " << caen_error (err);

  CAEN_DGTZ_AcqMode_t acquisition_mode = cma::get (acquisition_modes, ps.get<std::string> ("acquisition_mode", "HW"), "acquisition_mode");
  err = CAEN_DGTZ_SetAcquisitionMode (handle_, acquisition_mode);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetAcquisitionMode(" << handle_ << "," << cma::rget(acquisition_modes, acquisition_mode) << "): " << caen_error (err);

  if (ps.get<bool> ("count_all_triggers", true)) {
    set_register_bits (0x8100, 1 << 3);
    ev_offset_ = 0;
  } else ev_offset_ = 1;

  if (ps.get<bool> ("multiboard_sincronicazion", false)) {
    set_register_bits (0x8100, (1 << 8) | (1 << 5));  // bit 5 = almost busy set at #_buffers - 1
    set_register_bits (0x811C, 0xD0150);

    set_register (0x816C, 0);
    set_register (0x81A0, 0x2200);
  }

  sw_acquisition_ = (acquisition_mode == CAEN_DGTZ_SW_CONTROLLED);
  sw_trigger_ = ps.get<bool> ("sw_trigger", false);
  irqwait_ms_ = ps.get<int> ("irqwait_ms", 200);
  irqwait_checks_ = ps.get<int> ("irqwait_checks", 2);
}

void ds50::V172xDriver::init_buffers (ParameterSet const& ps) {
  int max_buffers_code = ps.get<int> ("max_buffers_code", 6);
  if (int(get_register(0x800C)) > max_buffers_code) set_register (0x800C, max_buffers_code);

  int max_events_blt = ps.get<int> ("max_events_blt", 20);
  CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_SetMaxNumEventsBLT (handle_, max_events_blt);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetMaxNumEventsBLT(" << handle_ << ","<< max_events_blt << "): " << caen_error (err);

  err = CAEN_DGTZ_SetInterruptConfig (handle_, CAEN_DGTZ_ENABLE, 1, 1, 1, CAEN_DGTZ_IRQ_MODE_RORA);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SetInterruptConfig(" << handle_ << ",CAEN_DGTZ_ENABLE,1,1,CAEN_DGTZ_IRQ_MODE_RORA): " << caen_error (err);

  err = CAEN_DGTZ_MallocReadoutBuffer (handle_, &event_buffer_, &buffer_size_);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_MallocReadoutBuffer(" << handle_ << "): " << caen_error (err);
}

void ds50::V172xDriver::start_ () {
  if (sw_acquisition_ || 1) {
    CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_SWStartAcquisition (handle_);
    if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SWStartAcquisition(" << handle_ << "): " << caen_error (err);
    usleep (10000);
  }

  statsHelper_.resetStatistics();
}

void ds50::V172xDriver::stop_ () {
  if (sw_acquisition_ || 1) {
    CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_SWStopAcquisition (handle_);
    if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SWStopAcquisition(" << handle_ << "): " << caen_error (err);
  }
}

bool ds50::V172xDriver::getNext__(artdaq::FragmentPtrs& frags) {
  bool will_stop = should_stop ();

  CAEN_DGTZ_ErrorCode err;
  if (sw_trigger_) {
    err = CAEN_DGTZ_SendSWtrigger (handle_);
    if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_SendSWtrigger(" << handle_ << "): " << caen_error (err);
  }

  artdaq::MonitoredQuantity::TIME_POINT_T startTime;
  for (int j = 0; j < irqwait_checks_; j++) {
    if (!(get_register (0x8104) & (1 << 3)) && !wait_irq ()) continue;

    startTime = artdaq::MonitoredQuantity::getCurrentTime();
    uint32_t gross_size;
    err = CAEN_DGTZ_ReadData (handle_, CAEN_DGTZ_SLAVE_TERMINATED_READOUT_MBLT, event_buffer_, &gross_size);
    if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_ReadData(" << handle_ << ",CAEN_DGTZ_SLAVE_TERMINATED_READOUT_MBLT," << event_buffer_ << "," << gross_size << "): " << caen_error (err);

    uint32_t n_events;
    err = CAEN_DGTZ_GetNumEvents (handle_, event_buffer_, gross_size, &n_events);
    if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_GetNumEvents(" << handle_ << "," << event_buffer_ << "," << gross_size << "," << n_events << "): " << caen_error (err);

    artdaq::MonitoredQuantity::TIME_POINT_T readTimePerFragment = 0.0;
    if (n_events > 0) {
      readTimePerFragment = (artdaq::MonitoredQuantity::getCurrentTime() - startTime) / n_events;
    }

    for (uint32_t i = 0; i < n_events; i++) {
      statsHelper_.addSample(READ_TIME_STAT_KEY, readTimePerFragment);
      CAEN_DGTZ_EventInfo_t event_info;
      char *event_ptr;
      err = CAEN_DGTZ_GetEventInfo (handle_, event_buffer_, gross_size, i, &event_info, &event_ptr);
      if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_GetEventInfo(" << handle_ << "," << event_buffer_ << "," << gross_size << "," << i << "," << event_info.EventCounter << "," << event_ptr << "): " << caen_error (err);

      uint32_t ev_counter_masked = (ev_counter () - ev_offset_) & 0xFFFFFFUL;
      uint32_t ev_counter_board = reinterpret_cast<const V172xFragment::Header *>(event_ptr)->event_counter;
      int ev_counter_diff = ev_counter_board - ev_counter_masked;
      if (ev_counter_diff > 0) {
        mf::LogError ("V172x") << "board " << fragment_id () << " jump " << ev_counter_diff << " at event " << ev_counter ();

        for (int i = 0; generate_fragment_ && ev_counter_diff < 100 && i < ev_counter_diff; ++i)
          frags.emplace_back(new artdaq::Fragment(ev_counter () + i, fragment_id (), Config::MISSED_FRAGMENT_TYPE));

        ev_counter_inc (ev_counter_diff);
      } else if (ev_counter_diff < 0) {
        mf::LogError ("V172x")  << "board " << fragment_id () << " duplicated fragment " << ev_counter_board << " at event " << ev_counter ();
        continue;
      }


      uint32_t pattern = reinterpret_cast<const V172xFragment::Header *>(event_ptr)->pattern & 0xff;
      if (pattern_check_ && (pattern != (ev_counter () & 0xff)))
        mf::LogError ("V172x") << "board " << fragment_id () << " pattern mismatch " << std::hex << std::showbase << pattern << " vs " << (ev_counter () & 0xff) << " at event " << std::dec << ev_counter ();


      if (generate_fragment_) {
        startTime = artdaq::MonitoredQuantity::getCurrentTime();
        std::unique_ptr<artdaq::Fragment> frag(new artdaq::Fragment(std::ceil (event_info.EventSize / static_cast<double>(sizeof (artdaq::RawDataType))), ev_counter (), fragment_id (), fragment_type_, metadata_));
        artdaq::MonitoredQuantity::TIME_POINT_T endTime = artdaq::MonitoredQuantity::getCurrentTime();
        statsHelper_.addSample(FRAGMENT_CREATE_TIME_STAT_KEY, (endTime - startTime));

        startTime = endTime;
        memcpy (&*frag->dataBegin(), event_ptr, event_info.EventSize);
        statsHelper_.addSample(MEMCPY_TIME_STAT_KEY,
                              (artdaq::MonitoredQuantity::getCurrentTime() - startTime));

        frags.emplace_back (std::move (frag));
      }

      ev_counter_inc ();
      if (statsHelper_.readyToReport(READ_TIME_STAT_KEY, ev_counter())) {
        std::string statString = buildStatisticsString_();
        mf::LogDebug ("V172x") << statString;
      }
    }
  }

  if (!frags.size () && will_stop) return false;
  return true;
}

void ds50::V172xDriver::set_register (uint16_t reg, uint32_t val) {
  CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_WriteRegister (handle_, reg, val);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_WriteRegister(" << handle_ << "," << std::hex << std::showbase << reg << "," << val << "): " << caen_error (err);
}

uint32_t ds50::V172xDriver::get_register (uint16_t reg) {
  uint32_t val;

  CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_ReadRegister (handle_, reg, &val);
  if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_ReadRegister(" << handle_ << "," << std::hex << std::showbase << reg << "): " << caen_error (err);

  return val;
}

uint32_t ds50::V172xDriver::set_register_bits (uint16_t reg, uint32_t bits) {
  uint32_t val = get_register (reg);

  if (bits) {
    val |= bits;
    set_register (reg, val);
  }

  return val;
}

bool ds50::V172xDriver::wait_irq () {
  if (vme_) throw cet::exception("V172x") << "VME readout uninplemeted";
  CAEN_DGTZ_ErrorCode err = CAEN_DGTZ_IRQWait (handle_, irqwait_ms_);
  if (err == CAEN_DGTZ_Timeout) return false;
  else if (err != CAEN_DGTZ_Success) throw cet::exception("V172x") << "CAEN_DGTZ_IRQWait(" << handle_ << "," << irqwait_ms_ << "): " << caen_error (err);

  return true;
}

std::string ds50::V172xDriver::buildStatisticsString_()
{
  std::ostringstream oss;
  artdaq::MonitoredQuantityPtr mqPtr = artdaq::StatisticsCollection::getInstance().
    getMonitoredQuantity(READ_TIME_STAT_KEY);
  if (mqPtr.get() != 0) {
    artdaq::MonitoredQuantity::Stats stats;
    mqPtr->getStats(stats);
    oss << "Average times per fragment: read time = "
        << (stats.recentValueAverage) << " sec";
  }

  mqPtr = artdaq::StatisticsCollection::getInstance().
    getMonitoredQuantity(FRAGMENT_CREATE_TIME_STAT_KEY);
  if (mqPtr.get() != 0) {
    artdaq::MonitoredQuantity::Stats stats;
    mqPtr->getStats(stats);
    oss << ", Fragment creation time = "
        << stats.recentValueAverage << " sec";
  }

  mqPtr = artdaq::StatisticsCollection::getInstance().
    getMonitoredQuantity(MEMCPY_TIME_STAT_KEY);
  if (mqPtr.get() != 0) {
    artdaq::MonitoredQuantity::Stats stats;
    mqPtr->getStats(stats);
    oss << ", memcpy time = "
        << stats.recentValueAverage << " sec";
  }

  return oss.str();
}


DEFINE_ARTDAQ_GENERATOR(ds50::V172xDriver)
