#!/usr/bin/env ruby

require "open3"

def fileInUse(filePath)
  # Determine if a file is in use using the lsof utility.  If the file isn't in
  # use the utility will return no output to stdout.
  lsofResult = nil
  lsofError = nil
  Open3.popen3("/usr/sbin/lsof %s" % [filePath]) { |stdin, stdout, stderr|
    lsofResult = stdout.read
    lsofError = stderr.read
  }
  if lsofError != ""
    puts "lsof ERROR: " + lsofError
    return true
  end
  if lsofResult != ""
    return true
  end

  # 15-May-2013, KAB - adding double-check (since I saw an instance where a
  # current file was accidentally deleted).  In this test, we compare the
  # current time to the last modified time of the file.
  lastModifyTime = nil
  Open3.popen3("stat --format='%%Y' %s" % [filePath]) { |stdin, stdout, stderr|
    lastModifyTime = stdout.read
  }
  #puts lastModifyTime
  currentTime = nil
  Open3.popen3("date '+%s'") { |stdin, stdout, stderr|
    currentTime = stdout.read
  }
  #puts currentTime
  timeDiff = Integer(currentTime) - Integer(lastModifyTime)
  #puts timeDiff
  if timeDiff < 30
    return true
  end

  return false
end

def getMetaData(filePath)
  # Run the dataMetaData utility over the given file.  If the daqMetaData utility
  # doesn't return anything (either the file is empty or corrupt) return nil to
  # the caller.  If actual meta data is returned format it into a hash and
  # return it.
  output = nil
  Open3.popen3("daqMetaData %s" % [filePath]) { |stdin, stdout, stderr|
    output = stdout.read
  }

  lines = output.split("\n")
  if lines.length == 1
    return nil
  end

  firstEvent = lines[0].split("\t")
  lastEvent = lines[1].split("\t")

  return Hash["firstRun" => firstEvent[0], "lastRun" => lastEvent[0],
              "firstSubRun" => firstEvent[1], "lastSubRun" => lastEvent[1],
              "firstEvent" => firstEvent[2], "lastEvent" => lastEvent[2]]
end

def getUUID()
  # There is now UUID module in the version of ruby that we are using so we'll
  # use the command line utility.
  output = nil
  Open3.popen3("uuidgen") { |stdin, stdout, stderr|
    output = stdout.read
  }

  return output.chomp
end

if __FILE__ == $0
  if ARGV.length != 1
    puts "Usage:"
    puts "  %s /path/to/data/directory" % [__FILE__]
    exit
  end

  fileEntries = Dir.entries(ARGV[0])
  fileEntries.each do |fileEntry|
    # Assume that the DAQ writes out files starting with "ds50daq"
    # or "RootOutput" and ending with "root"
    if fileEntry =~ /^ds50daq.*root$/ || fileEntry =~ /^RootOutput.*root$/
      fullPath = File.join(ARGV[0], fileEntry)
      if fileInUse(fullPath)
        puts "Skipping %s, in use." % [fileEntry]
        next
      end

      metaData = getMetaData(fullPath)
      if metaData == nil
        #puts "Deleting %s" % [fileEntry]
        #File.delete(fullPath)

        # 14-May-2013, KAB - Let's rename files that we suspect to
        # be bad rather than delete them (for safety).
        puts "File %s appears to be empty or unreadable" % [fileEntry]
        newName = "%s.empty" % fileEntry
        puts "Renaming %s to %s" % [fileEntry, newName]
        File.rename(fullPath, File.join(ARGV[0], newName))
      else
        # I think it's a good idea to include a UUID in each file as I can think
        # of valid use cases where we'd want to duplicate events to different
        # files and run-subrun-event wouldn't be a unique identifier.
        #newName = "ds50data_run%s_subrun%s_event%s-run%s_subrun%s_event%s-%s.root" % [metaData["firstRun"],
        #                                                                              metaData["firstSubRun"],
        #                                                                              metaData["firstEvent"],
        #                                                                              metaData["lastRun"],
        #                                                                              metaData["lastSubRun"],
        #                                                                              metaData["lastEvent"],
        #                                                                              getUUID()]

        # 14-May-2013, KAB - implementing filename pattern that was discussed
        # in the control room this morning.
        newPrefix = "ds50_r%06d_sr%02d_" % [metaData["firstRun"], metaData["firstSubRun"]]
        newName = fileEntry.sub("ds50daq_", newPrefix)
        puts "Renaming %s to %s" % [fileEntry, newName]
        File.rename(fullPath, File.join(ARGV[0], newName))
      end
    end
  end
end
