#include "cetlib/exception.h"
#include "ds50daq/online/ArtServices/ds50db_interface.hh"

ds50::db::interface::interface (const std::string& host, const std::string& db, const std::string& user, const std::string& pwd): current_run_(-1) {
  std::string connection_info = "host = " + host + " user = " + user + " password = " + pwd + " dbname = " + db;
  try {
    c_ = std::unique_ptr<connection>(new connection (connection_info));
  } catch (const std::runtime_error& e) {
    throw cet::exception("dbi::dbi") << "can not connect: " << e.what ();
  }
}

int ds50::db::interface::latest_run (const std::string& table_, const std::string& version) {
  std::string q_s = "SELECT run FROM " + table_ + " WHERE run < " + quote (current_run_);
  if (version.size ()) q_s += " AND version = " + quote (version);
  q_s += "ORDER BY run DESC LIMIT 1";
  result r = query (q_s);
  if (r.empty ()) return 0;
  return r.get<int> (0, 0);
}

ds50::db::result ds50::db::interface::latest (const std::string& table, const std::string& version) {
  int run = latest_run (table, version);
  return query ("SELECT * FROM " + table + " WHERE run = " + quote (run) + (version.size () ? (" AND version = " + quote (version)) : "")).set_table_name (table);
}

ds50::db::result ds50::db::interface::current (const std::string& table, const std::string& version) {
  return query ("SELECT * FROM " + table + " WHERE run = " + quote (current_run_) + (version.size () ? (" AND version = " + quote (version)) : "")).set_table_name (table);
}

ds50::db::result ds50::db::interface::insert (const ds50::db::result& r, const std::string& version) {
  if (!version.size ()) throw cet::exception("dbi::insert") << "version argument is required";
  if (!r.table_name ().size ()) throw cet::exception("dbi::insert") << "table name unset (probably not the output of a simple query)";
  for (size_t i = 0; i < r.rows (); i++) if (r.get<std::string> ("version", i) != version) throw cet::exception("dbi::insert") << "version in table " << r.get<std::string> ("version", i) << ") does not agree with the specified " + version;
  auto column_names = r.column_names ();
  std::string q = "INSERT INTO " + r.table_name () + " (";
  for (const auto& name: column_names) q += name + ",";
  q[q.size () - 1] = ')';
  q += " VALUES ";

  for (size_t row = 0; row < r.rows (); row++) {
    q += "(";
    for (const auto& name: column_names) q += quote(r.get<std::string>(name, row)) + ",";
    q[q.size () - 1] = ')';
    q += ",";
  }
  q[q.size () - 1] = ';';

  return query(q, true).set_table_name ("insert_results");
}

ds50::db::result ds50::db::interface::query (const std::string& q, bool autocommit) { 
  try {
    return transaction (autocommit)->exec (q); 
  } catch (const pqxx::broken_connection& e) {
    throw cet::exception("dbi::query") << "broken connection: " << e.what ();
  } catch (const pqxx::undefined_table& e) {
    throw cet::exception("dbi::query") << "undefined table: " << e.what ();
  } catch (const pqxx::syntax_error& e) {
    throw cet::exception("dbi::query") << "syntax error: " << e.what ();
  } catch (const pqxx::insufficient_privilege& e) {
    throw cet::exception("dbi::query") << "insufficient privilege: " << e.what ();
  } catch (const pqxx::sql_error& e) {
    throw cet::exception("dbi::query") << "sql error: " << e.what ();
  } catch (...) {
    throw cet::exception("dbi::query") << "unknown exception";
  }
}

std::unique_ptr<ds50::db::interface::work> ds50::db::interface::transaction (bool autocommit) { return std::unique_ptr<work>(new work(*c_, autocommit)); }

ds50::db::interface::work::work (connection& c, bool autocommit): w_(c), lk_(mutex_), autocommit_(autocommit) {}
ds50::db::interface::work::~work () { if (autocommit_) w_.commit (); }
ds50::db::result ds50::db::interface::work::exec (const std::string& q) { return result(w_.exec(q)); }
void ds50::db::interface::work::commit () { w_.commit (); }
