#include "art/Persistency/Provenance/FileIndex.h"

#include <boost/filesystem.hpp>

#include "Cintex/Cintex.h"
#include "TFile.h"
#include "TTree.h"

#include <iostream>

struct DataHandles {
  TTree *metaDataTree;
  TFile *artFile;
  art::FileIndex *fileIndex;
};

void cleanup(struct DataHandles *dataHandles) {
  /* Cleanup needs to be done manually and in the correct order otherwise we
     get crashes and hangs when the program exists. */
  if (dataHandles->metaDataTree) {
    dataHandles->metaDataTree->SetBranchAddress("FileIndex", 0);
  }

  if (dataHandles->fileIndex) {
    delete dataHandles->fileIndex;
  }

  dataHandles->fileIndex = 0;

  if (dataHandles->metaDataTree) {
    delete dataHandles->metaDataTree;
  }

  dataHandles->metaDataTree = 0;

  if (dataHandles->artFile) {
    dataHandles->artFile->Close();
    delete dataHandles->artFile;
  }

  dataHandles->artFile = 0;
  delete dataHandles;
  return;
}

int main(int argc, char **argv) {
  /* This program will attempt to read a ROOT file written out by art and
     extract the FileIndex object which contains the run/subrun/event metadata.
     It will find the first and last event written.  That information will be
     written to stdout as two tab delimited lines as follows:
       first run\tfirst subrun\tfirst event\n
       last run\tlast subrun\tlast event\t

     If the file is corrupt or empty the following will be written to stdout:
       empty\tempty\tempty\n
  */
  ROOT::Cintex::Cintex::Enable();

  if (argc != 2) {
    return -1;
  }

  if (!boost::filesystem::exists(argv[1])) {
    std::cout << "empty\tempty\tempty" << std::endl;
    return -1;
  }

  struct DataHandles *dataHandles = new struct DataHandles;
  dataHandles->metaDataTree = 0;
  dataHandles->artFile = new TFile(argv[1]);
  dataHandles->fileIndex = new art::FileIndex();

  dataHandles->artFile->GetObject("MetaData", dataHandles->metaDataTree);
  if (dataHandles->metaDataTree == 0) {
    std::cout << "empty\tempty\tempty" << std::endl;
    cleanup(dataHandles);
    return -1;
  }

  dataHandles->metaDataTree->SetBranchAddress("FileIndex", &dataHandles->fileIndex);
  Long64_t numEntries = dataHandles->metaDataTree->GetEntries();
  if (numEntries != 1) {
    std::cout << "empty\tempty\tempty" << std::endl;
    cleanup(dataHandles);
    return -1;
  }

  Long64_t bytesRead = dataHandles->metaDataTree->GetEntry(0);
  if (bytesRead == 0) {
    std::cout << "empty\tempty\tempty" << std::endl;
    cleanup(dataHandles);
    return -1;
  }

  dataHandles->fileIndex->sortBy_Run_SubRun_Event();
  std::vector<art::FileIndex::Element>::const_iterator firstElement = dataHandles->fileIndex->begin();
  std::vector<art::FileIndex::Element>::const_iterator lastElement = dataHandles->fileIndex->end() - 1;
  
  if (dataHandles->fileIndex->begin() == dataHandles->fileIndex->end()) {
    std::cout << "empty\tempty\tempty" << std::endl;
    cleanup(dataHandles);
    return -1;
  }    

  bool foundEvent = false;
  while(!foundEvent && firstElement != dataHandles->fileIndex->end()) {
    if (firstElement->eventID_.runID().isValid() && 
	firstElement->eventID_.subRunID().isValid() &&
	firstElement->eventID_.isValid()) {
      foundEvent = true;
    } else {
      ++firstElement;
    }
  }

  if (!foundEvent) {
    std::cout << "empty\tempty\tempty" << std::endl;
    cleanup(dataHandles);
    return -1;
  }

  foundEvent = false;
  while(!foundEvent && lastElement != dataHandles->fileIndex->begin()) {
    if (lastElement->eventID_.runID().isValid() &&
	lastElement->eventID_.subRunID().isValid() &&
	lastElement->eventID_.isValid()) {
      foundEvent = true;
    } else {
      --lastElement;
    }
  }

  if (!foundEvent) {
    std::cout << "empty\tempty\tempty" << std::endl;
    cleanup(dataHandles);
    return -1;
  }

  std::cout << firstElement->eventID_.run() << "\t" << firstElement->eventID_.subRun() << "\t" << firstElement->eventID_.event() << std::endl;
  std::cout << lastElement->eventID_.run() << "\t" << lastElement->eventID_.subRun() << "\t" << lastElement->eventID_.event() << std::endl;

  cleanup(dataHandles);
  return 0;
}
